unit GWVisio_Controller;
{---------------------------------------------------------
Visio Control Unit
------------------

Implements TGWVisio_Controller object that collects together functions
for connecting to Visio (ie: Visio OLE Automation server) and for
responding to events.

Before compiling this unit into a project, you will need to
generate the Visio_TLB.pas file. In D4, simply use
Project -- Import... and import the Visio type library.

----------------------------
Version History
----------------------------
99-01-30: GW minor fixups
99-01-25: GW added DockedStencil function
99-01-18: GW added utility functions for helping Master.Open/Close process
99-01-07: GW added flag for "call refused" tracking
98-11-02: GW original

----------------------------
Graham Wideman
graham@wideman-one.com
---------------------------

---------------------------------------------------------}
interface

uses ComObj, ActiveX, Classes,
     Visio_TLB,          // so we can call Visio
     GWVisio_AddOn_TLB   // so we can "inherit" the declaration of IVisioAddOnSink
     ;

{$R GWVisio_AddOn.TLB}  // include TLB resource for IVisioAddOnSink

//--------------------------------------------------------
// Define some constants
//--------------------------------------------------------
Const
  VisioClassName   = 'visio.application';
  ECallRefusedCode = -2147418111;

  // tvoc_xxx Visio Connect options
  tvco_None          = 0;
  tvco_New           = 1;  // open new instance of Visio only
  tvco_Existing      = 2;  // connect only to running instance of Visio
  tvco_TryExisting   = tvco_New OR tvco_Existing;  // try existing first, else create new

Type
  // vcr_xxx: Visio Controller Result
  Tvcr_xxx = (
    vcr_OK,
    vcr_OK_NoContainingMaster,
    vcr_Err_MasterOpenFailed,
    vcr_Err_MasterCloseFailed,
    vcr_Err_SetValueFailed,
    vcr_CallRefused,
    vcr_Err_Misc
  );

Const
  vcr_Begin = vcr_OK;
  vcr_End   = vcr_Err_Misc;
  vcr_Captions: array[vcr_Begin..vcr_End] of string = (
     'OK',
     'OK - No containing Master',
     'Error - Master.Open failed',
     'Error - Master.Close failed',
     'Error - Set value failed',
     'Error - Call refused by Visio',
     'Error - Misc'
    );

Type
  //---------------------------
  // Some forward declarations
  //---------------------------
  TVisioAddOnSink    = class;
  TGWVisio_EventInfo = class;

  //------------------------------------------------------------------
  // User-supplied event-handler function, to delegate  event-response
  // to user code. Just like VCL way of doing things.
  //------------------------------------------------------------------
  TVisEvent = function (AGWVisio_EventInfo: TGWVisio_EventInfo): HResult of object;
  // Note: "of object" means function must be a member of an object, not a free-floating function.

//**********************************************************************
// Visio Controller: This is the heart of this unit
//**********************************************************************
  //--------------------------
  TGWVisio_Controller = class
  //--------------------------
    //-------------------------
    // Visio connection
    //-------------------------
    private
      FApplication            : IVApplication;
    public
      property Application: IVApplication read FApplication;
      constructor Create; virtual;
      destructor  Destroy; override;

      function Connect(AOptions: integer): HResult;
      function Connected: boolean;
      function Disconnect: HResult;

    //-------------------------
    // Event Handling
    //-------------------------
    private
      FVisioAddOnSink         : IVisioAddOnSink;
      Pas_VisioAddOnSink      : TVisioAddOnSink;
      FEventRegistrations     : array of IVEvent;  // maintain them here so we can delete them later
      FEventRegistrationsCount : integer;
      function GetEventRegistrations(index: integer): IVEvent;
    public
      property EventRegistrations[index: integer]: IVEvent read GetEventRegistrations;
      property EventRegistrationsCount : integer read FEventRegistrationsCount;

      function EventHandling_Init(AOnVisEvent: TVisEvent { callback proc }): HResult;
      procedure TestOnVisEvent;
      function EventHandling_Disconnect: HResult;

      function RegisterForEvent(
                 WatchObject  : OleVariant;   // IVxxx object whose events to register for
                 AEvtxxx      : longword;     // see Visio ProgRef.hlp on event codes...
                 targetArgs   : string
               ): integer;                    // Returns index of IVEvent in EventRegistrations;
                                              // or -1 for unable to register

      procedure UnregisterEvent(EventNum: integer);

    //------------------------------
    // Utility: Docked Stencil array
    //------------------------------
    private
      FStencilArray          : OleVariant;

    public
      function GetDockedStencilList(Win: IVWindow; var SL: TStringList; var Msg: string): Tvcr_xxx;

    //-----------------------------------------------------------------
    // Helper functions for IVMaster.Open/Close
    //-----------------------------------------------------------------
    // Useful for modifying individual Master Shape or Cell properties.
    // Given a Shape or Cell, finds the containing Master, clones it via
    // Open, then works back "down" to find the cloned equivalent of the
    // original Shape or Cell.
    //-----------------------------------------------------------------
    public
      function FindMasterTemp(MasterIn      : IVMaster;
                         out  MasterObjTemp : IVMaster;
                         out  MasterTemp    : IVMaster;
                         var  Msg           : string
                         ): Tvcr_xxx;

      function FindShapeTemp( ShapeIn      : IVShape;
                         out  ShapeTemp    : IVShape;
                         out  MasterTemp   : IVMaster;
                         var  Msg           : string
                         ): Tvcr_xxx;

      function FindCellTemp(   CellIn      : IVCell;
                         out  CellTemp     : IVCell;
                         out  MasterTemp   : IVMaster;
                         var  Msg           : string
                         ): Tvcr_xxx;

    //------------------------------------------
    // Utility: Last-call-refused (experimental)
    //------------------------------------------
    private
      FLastCallRefused      : Boolean;
      procedure SetLastCallRefused(B: Boolean);
    public
      LastCall_Callback     : procedure of object;
      property LastCallRefused : Boolean read FLastCallRefused write SetLastCallRefused;
  end;


  //------------------------------------------------------------------
  // Object for event info to hand over to user app to describe Event
  //------------------------------------------------------------------
  //------------------------
  TGWVisio_EventInfo = class
  //------------------------
  private
    FEvent     : Smallint   ;
    FSource    : IUnknown   ;
    FID        : integer    ;
    FSeq       : Integer    ;
    FSubject   : IUnknown   ;
    FExtra     : OleVariant ;
  public
    property Event     : Smallint   read FEvent    ;  // really "event id"
    property Source    : IUnknown   read FSource   ;
    property ID        : integer    read FID       ;
    property Seq       : Integer    read FSeq      ;
    property Subject   : IUnknown   read FSubject  ;
    property Extra     : OleVariant read FExtra    ;

    constructor Create(
        AEvent     : Smallint   ;
        ASource    : IUnknown   ;
        AID        : integer    ;
        ASeq       : Integer    ;
        ASubject   : IUnknown   ;
        AExtra     : OleVariant
      );
    // destructor -- default is fine.
  end;

  //-----------------------------------------------------------
  // TVisioAddOnSink implements that IVisioAddOnSink interface
  // that we must pass to Visio for Visio to talk to when it sends
  // event notifications.
  //-----------------------------------------------------------

  //---------------------------------------------------------
  TVisioAddOnSink = class(TAutoObject, IVisioAddOnSink)
  //---------------------------------------------------------
  protected
    // VisEventProc implements the proc in IVisioAddOnSink that
    // Visio wants to know about.
    procedure VisEventProc(
                        Event: Smallint;
                 const Source: IUnknown;
                           ID: integer;
                          Seq: Integer;
                const Subject: IUnknown;
                        Extra: OleVariant
                        ); safecall;
    private
      FOnVisEvent: TVisEvent;
    public
      property    OnVisEvent: TVisEvent read FOnVisEvent write FOnVisEvent;
      constructor Create(AVisEvent: TVisEvent);  // supply event handler function
  end;


Var
  //------------------------------------------------------------
  // Singleton Visio_Controller object. See unit initialization.
  //------------------------------------------------------------
  Visio_Controller: TGWVisio_Controller;

//----------------------------
implementation
//----------------------------
uses Windows, ComServ, SysUtils;

//------------------------------------------------------------------
//  TVisioAddOnSink
//------------------------------------------------------------------

//------------------------------------------------------------------
constructor TVisioAddOnSink.Create(AVisEvent: TVisEvent);
//------------------------------------------------------------------
Begin
  inherited Create;
  FOnVisEvent := AVisEvent;
end;

//------------------------------------------------------------------
procedure  TVisioAddOnSink.VisEventProc(
//------------------------------------------------------------------
          Event   : Smallint;
    const Source  : IUnknown;
          ID      : integer;
          Seq     : Integer;
    const Subject : IUnknown;
          Extra   : OleVariant
          );
Var
  AEventInfo: TGWVisio_EventInfo;
  hr: HResult;
begin

  If Assigned(FOnVisEvent) then
  Begin
    AEventInfo := TGWVisio_EventInfo.Create(
                  Event   ,
                  Source  ,
                  ID      ,
                  Seq     ,
                  Subject ,
                  Extra   );

    hr := FOnVisEvent(AEventInfo);
  end;
end;

//------------------------------------------------------------
//          TGWVisio_EventInfo
//------------------------------------------------------------

//------------------------------------------------------------------
constructor TGWVisio_EventInfo.Create(
//------------------------------------------------------------------
        AEvent     : Smallint   ;
        ASource    : IUnknown   ;
        AID        : integer    ;
        ASeq       : Integer    ;
        ASubject   : IUnknown   ;
        AExtra     : OleVariant);
Begin
  FEvent     :=  AEvent     ;
  FSource    :=  ASource    ;
  FID        :=  AID        ;
  FSeq       :=  ASeq       ;
  FSubject   :=  ASubject   ;
  FExtra     :=  AExtra     ;
end;

//------------------------------------------------------------
//          TGWVisio_Controller
//------------------------------------------------------------

//------------------------------------------------------------
constructor TGWVisio_Controller.Create;
//------------------------------------------------------------
Begin
  FApplication       := nil;
  FVisioAddOnSink    := nil;
  Pas_VisioAddOnSink := nil;
  LastCallRefused   := false;
  FStencilArray := VarArrayCreate([0, 0], varOleStr); // initialize an empty VarArray structure

  // EventRegistrations
  FEventRegistrationsCount := 0;

end;

//------------------------------------------------------------
destructor TGWVisio_Controller.Destroy;
//------------------------------------------------------------
Begin
  // nil everything just in case.
  FApplication         := nil;
  FVisioAddOnSink      := nil;
  Pas_VisioAddOnSink   := nil;
  FStencilArray        := UnAssigned;
  inherited Destroy;
end;

//------------------------------------------------------------
function TGWVisio_Controller.Connect(AOptions: integer): HResult;
//------------------------------------------------------------
Var
  MyIDispatch: IDispatch;
Label connected;
begin
  Result := E_FAIL;

  If Assigned(FApplication) then Disconnect;

  If (AOptions AND tvco_Existing) <> 0 then
  Begin  // try to connect to existing Visio instance
    try
      MyIDispatch := GetActiveOLEObject(VisioClassName);
      FApplication  := MyIDispatch as IVApplication;
    except
      ; // do nothing except catch exception for no instance running
    end;
  end;

  If (not Assigned(FApplication)) and ((AOptions AND tvco_New) <> 0) then
  Begin  // start new instance
    try
      FApplication := Visio_TLB.CoApplication_.Create;
    except
      ; // do nothing except catch exceptions
    end;
  end;
connected:
  if Assigned(FApplication) then
  Begin
    Result           := NOERROR;
    LastCallRefused  := false;
  end;
end;

//------------------------------------------------------------
function TGWVisio_Controller.Connected: boolean;
//------------------------------------------------------------
Begin
  Result := Assigned(Application);
end;

//------------------------------------------------------------
function TGWVisio_Controller.Disconnect: HResult;
//------------------------------------------------------------
begin
  EventHandling_Disconnect;
  FApplication := nil;
  Result := NOERROR;
end;

//------------------------------------------------------------
function TGWVisio_Controller.EventHandling_Init(AOnVisEvent: TVisEvent): HResult;
//------------------------------------------------------------
begin
  Result := E_FAIL;
  FVisioAddOnSink    := nil;
  Pas_VisioAddOnSink := nil;

  try
    Pas_VisioAddOnSink := TVisioAddOnSink.Create(AOnVisEvent);
    if Assigned(Pas_VisioAddOnSink) then
      FVisioAddOnSink := Pas_VisioAddOnSink as IVisioAddOnSink;
  except
  end;
  if Assigned(FVisioAddOnSink) then Result := NOERROR;
end;

//------------------------------------------------------------
function TGWVisio_Controller.EventHandling_Disconnect: HResult;
//------------------------------------------------------------
Var
  EventNum: integer;
begin
  // first disable VisioAddOnSink's ability to callback
  if Assigned(Pas_VisioAddOnSink) then Pas_VisioAddOnSink.OnVisEvent := nil;
  // now get rid of our reference, so Visio is able cause dispose
  FVisioAddOnSink     := nil;

  For EventNum := 0 to EventRegistrationsCount - 1 do UnregisterEvent(EventNum);

  Pas_VisioAddOnSink  := nil;
  Result := NOERROR;
end;

//------------------------------------------------------------
function TGWVisio_Controller.RegisterForEvent(
               WatchObject: OleVariant;
               AEvtxxx    : longword;
               targetArgs : string
               ): integer;
//------------------------------------------------------------
Var
  visEventList : IVEventList;
  visEvent     : IVEvent;
  varSink      : OleVariant;
  iunk         : IUnknown;
  EventNum     : integer;
Label the_end, done_1;
begin
  Result       := -1;
  visEventList := nil;

  if not Assigned(FVisioAddOnSink) then goto the_end;

  // Try to get an event list
  try
    iunk          := WatchObject.EventList;
    visEventList  := iunk as IVEventList;
  except
  end;
  if not Assigned(visEventList) then goto the_end;

  varSink := OleVariant(FVisioAddOnSink);
  visEvent  := visEventList.AddAdvise(
               AEvtxxx,
               OleVariant(FVisioAddOnSink),
               WideString(''),
               WideString(targetArgs));

  // try to find a blank spot
  For EventNum := 0 to FEventRegistrationsCount - 1 do
  Begin
    If not Assigned(FEventRegistrations[EventNum]) then
    Begin
      FEventRegistrations[EventNum] := visEvent;
      Result := EventNum;
      goto done_1;
    end;
  end;
  // append to array
  If FEventRegistrationsCount > High(FEventRegistrations) then
    SetLength(FEventRegistrations, High(FEventRegistrations) + 5);
  FEventRegistrations[FEventRegistrationsCount] := visEvent;
  Result := FEventRegistrationsCount;
  Inc(FEventRegistrationsCount);
done_1:
the_end:
end;

//------------------------------------------------------------
procedure TGWVisio_Controller.UnregisterEvent(EventNum: integer);
//------------------------------------------------------------
begin
  If Assigned(EventRegistrations[EventNum]) then
  Begin
    try
      EventRegistrations[EventNum].Delete;
    except
    end;
    FEventRegistrations[EventNum] := nil;
  end;
end;

//------------------------------------------------------------
procedure TGWVisio_Controller.TestOnVisEvent;
//------------------------------------------------------------
Begin
  Pas_VisioAddOnSink.OnVisEvent(nil);
end;

//------------------------------------------------------------
procedure TGWVisio_Controller.SetLastCallRefused(B: Boolean);
//------------------------------------------------------------
Var
  OldLastCallRefused: boolean;
begin
  OldLastCallRefused := FLastCallRefused;
  FLastCallRefused := B;
  if Assigned(LastCall_Callback) and
    (OldLastCallRefused <> FLastCallRefused) then LastCall_Callback;
end;

//--------------------------------------------------------
function TGWVisio_Controller.FindMasterTemp(
                            MasterIn      : IVMaster;
                       out  MasterObjTemp : IVMaster;
                       out  MasterTemp    : IVMaster;
                       var  Msg           : string
                       ): Tvcr_xxx;
//--------------------------------------------------------
Begin
  Result        := vcr_OK;
  Msg           := '';
  MasterObjTemp := nil;
  MasterTemp    := nil;
  try
    MasterTemp     :=  MasterIn.Open;
    MasterObjTemp  :=  MasterTemp;
    Result := vcr_OK;
  except
    on E: Exception do
    Begin
      Msg := ' {'+E.Message+'}';
      Result := vcr_Err_MasterOpenFailed;
    end;
  end;
end;

//--------------------------------------------------------
function TGWVisio_Controller.FindShapeTemp(
                            ShapeIn      : IVShape;
                       out  ShapeTemp    : IVShape;   // return
                       out  MasterTemp   : IVMaster;
                       var  Msg           : string
                       ): Tvcr_xxx;
//--------------------------------------------------------
Var
  Master: IVMaster;
  ContainingShape, ContainingShapeTemp: IVShape;
  index: integer;
  TempResult: Tvcr_xxx;
  TempMsg: string;
Begin
  Result := vcr_OK;
  Msg := '';
  ShapeTemp := nil;
  MasterTemp := nil;
  try
    Master := ShapeIn.ContainingMaster;
    If Assigned(Master) then
    Begin  // it's worth doing...
      ContainingShape := ShapeIn.ContainingShape;
      index := ShapeIn.Index;
      If ContainingShape.Type_ = visTypePage then
      Begin // OK, ShapeIn is top group in Master.
        try
          MasterTemp := Master.Open;
        except
          on E: Exception do
          Begin
            Msg := Msg + ' {'+E.Message+'}';
            Result := vcr_Err_MasterOpenFailed;
          end;
        end;
        if Assigned(MasterTemp) then
        Begin
          ShapeTemp := MasterTemp.Shapes.Item[index];
          Result := vcr_OK;
        end;
      end else
      Begin // dig up a layer
        TempResult := FindShapeTemp(ContainingShape, ContainingShapeTemp, MasterTemp, TempMsg);
        If TempResult = vcr_OK then
        Begin
          ShapeTemp := ContainingShapeTemp.Shapes.Item[index];
          Result := vcr_OK;
        end else Msg := Msg + ' {' + TempMsg + '}';
      end;
    end else Result := vcr_OK_NoContainingMaster;
  except
    on E: Exception do
    Begin
      Msg := Msg + ' {'+E.Message+'}';
      Result := vcr_Err_Misc;
    end;
  end;
end;

//--------------------------------------------------------
function TGWVisio_Controller.FindCellTemp(
                             CellIn      : IVCell;
                       out  CellTemp     : IVCell;   // return
                       out  MasterTemp   : IVMaster;
                       var  Msg           : string
                       ): Tvcr_xxx;
//--------------------------------------------------------
Var
  Name, TempMsg: string;
  ContainingShape, ShapeTemp: IVShape;

Begin
  Result      := vcr_OK;
  Msg         := '';
  CellTemp    := nil;
  MasterTemp  := nil;
  try
    ContainingShape := CellIn.Shape;
    If Assigned(ContainingShape) then
    Begin  // it's worth doing...
      Name := CellIn.Name;
      Result := FindShapeTemp(ContainingShape, ShapeTemp, MasterTemp, TempMsg);
      If Result = vcr_OK then
      Begin
        CellTemp := ShapeTemp.Cells[Name];
      end else
      Begin
        if TempMsg <> '' then Msg := Msg + ' {'+TempMsg+'}';
      end;
    end else Result := vcr_OK_NoContainingMaster;
  except
    on E: Exception do
    Begin
      Msg := Msg + ' {'+E.Message+'}';
      Result := vcr_Err_Misc;
    end;
  end;
end;

//------------------------------------------------------------
function TGWVisio_Controller.GetDockedStencilList(
           Win   : IVWindow;
        var SL   : TStringList;
        var Msg  : string
                ): Tvcr_xxx;
//------------------------------------------------------------
Var
  MyPSafeArray: PSafeArray;
  SALo, SAHi, SANum: integer;
begin
  SL.Clear;

  try
    Win.DockedStencils(MyPSafeArray);                            // gives us a PSafeArray, we must dispose later
    PSafeArray(TVarData(FStencilArray).Varray)  := MyPSafeArray;  // give it to OLEVariant Array so we can manipulate easier
    SALo := VarArrayLowBound(FStencilArray,1);
    SAHi := VarArrayHighBound(FStencilArray,1);
    For SANum := SALo to SAHi do
    Begin
      SL.Add(FStencilArray[SANum]);
    end;
    SafeArrayDestroy(MyPSafeArray);
  except
  end;
end;

//------------------------------------------------------------
function TGWVisio_Controller.GetEventRegistrations(index: integer): IVEvent;
//------------------------------------------------------------
begin
  Result := FEventRegistrations[index];
end;

initialization
//------------------------------------------------------------
  TAutoObjectFactory.Create(ComServer,
                            TVisioAddOnSink, Class_VisioAddOnSink,
                            ciInternal, tmSingle);

  Visio_Controller := TGWVisio_Controller.Create;
end.

